<?php

namespace Drupal\aws_cloud\Controller\Ec2;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\aws_cloud\Service\AwsPricingServiceInterface;
use Drupal\aws_cloud\Aws\Ec2\InstanceTypePriceControllerInterface;

use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Controller responsible to show price list.
 */
class InstanceTypePriceController extends ControllerBase implements InstanceTypePriceControllerInterface {

  const ONE_YEAR = 365.25;

  /**
   * The Aws Pricing Service.
   *
   * @var \Drupal\aws_cloud\Service\AwsPricingServiceInterface
   */
  protected $awsPricingService;

  /**
   * The Messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Cloud context string.
   *
   * @var string
   */
  private $cloudContext;

  /**
   * Request object.
   *
   * @var \Symfony\Component\HttpFoundation\Request
   */
  private $request;

  /**
   * InstanceTypePriceController constructor.
   *
   * @param \Drupal\aws_cloud\Service\AwsPricingServiceInterface $aws_pricing_service
   *   AWS Pricing service.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   Messanger Object.
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   Request Object.
   */
  public function __construct(
    AwsPricingServiceInterface $aws_pricing_service,
    MessengerInterface $messenger,
    Request $request
  ) {
    $this->awsPricingService = $aws_pricing_service;
    $this->messenger = $messenger;
    $this->request = $request;
  }

  /**
   * Dependency Injection.
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('aws_cloud.pricing'),
      $container->get('messenger'),
      $container->get('request_stack')->getCurrentRequest()
    );
  }

  /**
   * {@inheritdoc}
   */
  public function show($cloud_context) {
    $this->cloudContext = $cloud_context;

    $build = [];

    $build['table'] = $this->buildTable();
    $build['#attached']['library'][] = 'aws_cloud/aws_cloud_instance_type_prices';

    return $build;
  }

  /**
   * Build table.
   *
   * @return array
   *   The render array of table.
   */
  private function buildTable() {
    $table = [
      '#type' => 'table',
      '#header' => array_values($this->getTableHeader()),
      '#attributes' => [
        'class' => ['aws_cloud_instance_type_prices'],
      ],
    ];

    foreach ($this->getTableRows() as $row) {
      $table[] = $row;
    }

    return $table;
  }

  /**
   * Get the header of table.
   *
   * @return array
   *   The header of table.
   */
  private function getTableHeader() {
    return [
      [
        'data' => $this->t('Instance Type'),
        'field' => 'instance_type',
      ],
      [
        'data' => $this->t('On-demand<br>Hourly ($)'),
        'field' => 'on_demand_hourly',
      ],
      [
        'data' => $this->t('On-demand<br>Daily ($)'),
        'field' => 'on_demand_daily',
      ],
      [
        'data' => $this->t('On-demand<br>Monthly ($)'),
        'field' => 'on_demand_monthly',
      ],
      [
        'data' => $this->t('On-demand<br>Yearly ($)'),
        'field' => 'on_demand_yearly',
      ],
      [
        'data' => $this->t('RI<br>1 Year ($)'),
        'field' => 'ri_one_year',
      ],
      [
        'data' => $this->t('RI<br>3 Year ($)'),
        'field' => 'ri_three_year',
      ],
    ];
  }

  /**
   * Get rows of table.
   *
   * @return array
   *   The render array of table's rows.
   */
  private function getTableRows() {
    $rows = [];
    $header = $this->getTableHeader();

    $rows_data = $this->getTableRowsData();
    foreach ($rows_data as $row_data) {
      $row = [];
      foreach ($row_data as $col_name => $col_val) {
        $this->buildCell($row, $col_name, $col_val);
      }
      $rows[] = $row;
    }

    return $rows;
  }

  /**
   * Get rows data of table.
   *
   * @return array
   *   The rows data of table.
   */
  private function getTableRowsData() {
    $instance_types = aws_cloud_get_instance_types($this->cloudContext);
    $rows_data = [];
    foreach ($instance_types as $key => $value) {
      $parts = explode(':', $value);
      $name = $parts[0];
      $hourly_rate = $parts[4];

      $rows_data[] = [
        'instance_type'     => $name,
        'on_demand_hourly'  => floatval($hourly_rate),
        'on_demand_daily'   => floatval($hourly_rate) * 24,
        'on_demand_monthly' => floatval($hourly_rate) * 24 * self::ONE_YEAR / 12,
        'on_demand_yearly'  => floatval($hourly_rate) * 24 * self::ONE_YEAR,
        'ri_one_year'       => floatval($parts[5]),
        'ri_three_year'     => floatval($parts[6]),
      ];
    }

    // Get sort and order parameters.
    $sort = $this->request->get('sort');
    $order_field = $this->getHeaderFieldByData($this->request->get('order'));
    if (empty($sort)) {
      $sort = 'asc';
    }
    if (empty($order_field)) {
      $order_field = 'instance_type';
    }

    // Sort rows data.
    usort($rows_data, function ($a, $b) use ($sort, $order_field) {
      if ($order_field == 'instance_type') {
        $a_type = explode('.', $a[$order_field])[0];
        $b_type = explode('.', $b[$order_field])[0];
        if ($a_type < $b_type) {
          $result = -1;
        }
        elseif ($a_type > $b_type) {
          $result = 1;
        }
        else {
          $result = $a['on_demand_hourly'] < $b['on_demand_hourly'] ? -1 : 1;
        }
      }
      else {
        $result = $a[$order_field] < $b[$order_field] ? -1 : 1;
      }

      if ($sort == 'desc') {
        $result *= -1;
      }

      return $result;
    });

    return $rows_data;
  }

  /**
   * Get field of header by data.
   *
   * @param string $data
   *   The data of header.
   *
   * @return string
   *   The field of header.
   */
  private function getHeaderFieldByData($data) {
    $field = NULL;
    $header = $this->getTableHeader();
    foreach ($header as $col) {
      if (is_array($col) && isset($col['data']) && $col['data'] == $data) {
        $field = $col['field'];
        break;
      }
    }
    return $field;
  }

  /**
   * Build cell.
   *
   * @param array &$row
   *   The render array of row.
   * @param string $cell_name
   *   The name of cell.
   * @param string $cell_value
   *   The value of cell.
   */
  private function buildCell(array &$row, $cell_name, $cell_value) {
    if ($cell_name == 'on_demand_hourly') {
      $precision = 4;
    }
    elseif ($cell_name == 'on_demand_daily' || $cell_name == 'on_demand_monthly') {
      $precision = 2;
    }
    else {
      $precision = 0;
    }

    if ($cell_name != 'instance_type') {
      $cell_value = $this->convertToNumber($cell_value, $precision);
    }

    $row[$cell_name] = [
      '#markup' => $cell_value,
    ];
  }

  /**
   * Convert to the string formatted with grouped thousands.
   *
   * @param float $float
   *   The float variable.
   * @param int $precision
   *   The optional number of decimal digits.
   *
   * @return string
   *   The formatted string.
   */
  private function convertToNumber($float, $precision = 0) {
    $float = round($float, $precision);
    return number_format($float, $precision);
  }

}
