<?php

namespace Drupal\aws_cloud\Service;

use Aws\Api\DateTimeResult;
use Aws\Credentials\AssumeRoleCredentialProvider;
use Aws\Credentials\CredentialProvider;
use Aws\Ec2\Ec2Client;
use Aws\Ec2\Exception\Ec2Exception;
use Aws\Endpoint\EndpointProvider;
use Aws\MockHandler;
use Aws\Result;
use Aws\Sts\StsClient;
use Drupal\aws_cloud\Entity\Ec2\ElasticIp;
use Drupal\aws_cloud\Entity\Ec2\Image;
use Drupal\aws_cloud\Entity\Ec2\Instance;
use Drupal\aws_cloud\Entity\Ec2\KeyPair;
use Drupal\aws_cloud\Entity\Ec2\NetworkInterface;
use Drupal\aws_cloud\Entity\Ec2\SecurityGroup;
use Drupal\aws_cloud\Entity\Ec2\Snapshot;
use Drupal\aws_cloud\Entity\Ec2\Volume;
use Drupal\cloud\Plugin\CloudConfigPluginManagerInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\core\Field\FieldTypePluginManagerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Messenger\Messenger;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslationInterface;

/**
 * AwsEc2Service service interacts with the AWS EC2 api.
 */
class AwsEc2Service implements AwsEc2ServiceInterface {

  use StringTranslationTrait;

  /**
   * The Messenger service.
   *
   * @var \Drupal\Core\Messenger\Messenger
   */
  protected $messenger;

  /**
   * Drupal\Core\Entity\EntityTypeManagerInterface definition.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Cloud context string.
   *
   * @var string
   */
  private $cloudContext;

  /**
   * A logger instance.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The config factory.
   *
   * Subclasses should use the self::config() method, which may be overridden to
   * address specific needs when loading config, rather than this property
   * directly. See \Drupal\Core\Form\ConfigFormBase::config() for an example of
   * this.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * CloudConfigPlugin.
   *
   * @var \Drupal\cloud\Plugin\CloudConfigPluginManagerInterface
   */
  protected $cloudConfigPluginManager;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * Field type plugin manager.
   *
   * @var \Drupal\core\Field\FieldTypePluginManagerInterface
   */
  protected $fieldTypePluginManager;

  /**
   * Entity field manager interface.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * TRUE to run the operation.  FALSE to run the operation in validation mode.
   *
   * @var bool
   */
  private $dryRun;

  /**
   * Constructs a new AwsEc2Service object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   An entity type manager instance.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   A logger instance.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   A configuration factory.
   * @param \Drupal\Core\Messenger\Messenger $messenger
   *   The messenger service.
   * @param \Drupal\Core\StringTranslation\TranslationInterface $string_translation
   *   The string translation service.
   * @param \Drupal\Core\Session\AccountInterface $current_user
   *   The current user.
   * @param \Drupal\cloud\Plugin\CloudConfigPluginManagerInterface $cloud_config_plugin_manager
   *   The cloud config plugin manager.
   * @param \Drupal\core\Field\FieldTypePluginManagerInterface $field_type_plugin_manager
   *   The field type plugin manager.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, LoggerChannelFactoryInterface $logger_factory, ConfigFactoryInterface $config_factory, Messenger $messenger, TranslationInterface $string_translation, AccountInterface $current_user, CloudConfigPluginManagerInterface $cloud_config_plugin_manager, FieldTypePluginManagerInterface $field_type_plugin_manager, EntityFieldManagerInterface $entity_field_manager) {
    // Setup the entity type manager for querying entities.
    $this->entityTypeManager = $entity_type_manager;
    // Setup the logger.
    $this->logger = $logger_factory->get('aws_ec2_service');
    // Setup the configuration factory.
    $this->configFactory = $config_factory;

    // Setup the dryRun flag.
    $this->dryRun = (bool) $this->configFactory->get('aws_cloud.settings')->get('aws_cloud_test_mode');

    // Setup the messenger.
    $this->messenger = $messenger;
    // Setup the $this->t()
    $this->stringTranslation = $string_translation;
    $this->currentUser = $current_user;
    $this->cloudConfigPluginManager = $cloud_config_plugin_manager;
    $this->fieldTypePluginManager = $field_type_plugin_manager;

    $this->entityFieldManager = $entity_field_manager;
  }

  /**
   * {@inheritdoc}
   */
  public function setCloudContext($cloud_context) {
    $this->cloudContext = $cloud_context;
    $this->cloudConfigPluginManager->setCloudContext($cloud_context);
  }

  /**
   * Load and return an Ec2Client.
   */
  private function getEc2Client() {
    $credentials = $this->cloudConfigPluginManager->loadCredentials();
    try {
      $ec2_params = [
        'region' => $credentials['region'],
        'version' => $credentials['version'],
      ];
      if ($credentials['use_instance_credentials'] == FALSE) {
        // Assume role using credential ini file.
        $provider = CredentialProvider::ini('default', $credentials['ini_file']);
        $provider = CredentialProvider::memoize($provider);

        if ($credentials['assume_role'] == TRUE) {

          $sts_params = [
            'region' => $credentials['region'],
            'version' => $credentials['version'],
            'credentials' => $provider,
          ];

          $assumeRoleCredentials = new AssumeRoleCredentialProvider([
            'client' => new StsClient($sts_params),
            'assume_role_params' => [
              'RoleArn' => $credentials['role_arn'],
              'RoleSessionName' => 'ec2_client_assume_role',
            ],
          ]);

          // Memoize takes care of re-authenticating when the tokens expire.
          $assumeRoleCredentials = CredentialProvider::memoize($assumeRoleCredentials);
          $ec2_params['credentials'] = $assumeRoleCredentials;
        }
        else {
          $ec2_params['credentials'] = $provider;
        }
      }
      $ec2_client = new Ec2Client($ec2_params);
    }
    catch (\Exception $e) {
      $ec2_client = NULL;
      $this->logger->error($e->getMessage());
    }
    $this->addMockHandler($ec2_client);
    return $ec2_client;
  }

  /**
   * Add a mock handler of aws sdk for testing.
   *
   * The mock data of aws response is saved
   * in configuration "aws_cloud_mock_data".
   *
   * @param \Aws\Ec2\Ec2Client $ec2_client
   *   The ec2 client.
   *
   * @see https://docs.aws.amazon.com/sdk-for-php/v3/developer-guide/guide_handlers-and-middleware.html
   */
  private function addMockHandler(Ec2Client $ec2_client) {
    $mock_data = $this->configFactory->get('aws_cloud.settings')->get('aws_cloud_mock_data');
    if ($this->dryRun && $mock_data) {
      $func = function ($command, $request) {
        $mock_data = \Drupal::service('config.factory')
          ->get('aws_cloud.settings')
          ->get('aws_cloud_mock_data');
        $mock_data = json_decode($mock_data, TRUE);

        // If the mock data of a command is defined.
        $command_name = $command->getName();
        if (isset($mock_data[$command_name])) {
          $result_data = $mock_data[$command_name];

          // Because launch time is special,
          // we need to convert it from string to DateTimeResult.
          if ($command_name == 'DescribeInstances') {
            foreach ($result_data['Reservations'] as &$reservation) {
              foreach ($reservation['Instances'] as &$instance) {
                if (!empty($instance['LaunchTime'])) {
                  $instance['LaunchTime'] = new DateTimeResult($instance['LaunchTime']);
                }
              }
            }
          }

          return new Result($result_data);
        }
        elseif ($command_name == 'DescribeAccountAttributes') {
          // Return an empty array so testing doesn't error out.
          $result_data = [
            'AccountAttributes' => [
              [
                'AttributeName' => 'supported-platforms',
                'AttributeValues' => [
                  [
                    'AttributeValue' => 'VPC',
                  ],
                ],
              ],
            ],
          ];
          return new Result($result_data);
        }
        else {
          return new Result();
        }
      };

      // Set mock handler.
      $ec2_client->getHandlerList()->setHandler(new MockHandler([$func, $func]));
    }
  }

  /**
   * Execute the API of AWS EC2 Service.
   *
   * @param string $operation
   *   The operation to perform.
   * @param array $params
   *   An array of parameters.
   *
   * @return array
   *   Array of execution result or NULL if there is an error.
   *
   * @throws \Drupal\aws_cloud\Service\AwsEc2ServiceException
   *   If the $params is empty or Ec2 Client is null.
   */
  private function execute($operation, array $params = []) {
    $results = NULL;

    if (empty($params)) {
      throw new AwsEc2ServiceException(sprintf("No parameters passed for operation %s", ['%s' => $operation]));
    }

    $ec2_client = $this->getEc2Client();
    if ($ec2_client == NULL) {
      throw new AwsEc2ServiceException('No Ec2 Client found.  Cannot perform API operations');
    }

    try {
      // Let other modules alter the parameters
      // before they are sent through the API.
      \Drupal::moduleHandler()->invokeAll('aws_cloud_pre_execute_alter', [
        &$params,
        $operation,
        $this->cloudContext,
      ]);

      $command = $ec2_client->getCommand($operation, $params);
      $results = $ec2_client->execute($command);

      // Let other modules alter the results before the module processes it.
      \Drupal::moduleHandler()->invokeAll('aws_cloud_post_execute_alter', [
        &$results,
        $operation,
        $this->cloudContext,
      ]);
    }
    catch (Ec2Exception $e) {
      $this->messenger->addError($this->t('Error: The operation "@operation" could not be performed.', [
        '@operation' => $operation,
      ]));

      $this->messenger->addError($this->t('Error Info: @error_info', [
        '@error_info' => $e->getAwsErrorCode(),
      ]));

      $this->messenger->addError($this->t('Error from: @error_type-side', [
        '@error_type' => $e->getAwsErrorType(),
      ]));

      $this->messenger->addError($this->t('Status Code: @status_code', [
        '@status_code' => $e->getStatusCode(),
      ]));

      $this->messenger->addError($this->t('Message: @msg', ['@msg' => $e->getAwsErrorMessage()]));

    }
    catch (\InvalidArgumentException $e) {
      $this->messenger->addError($e->getMessage());
    }
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function associateAddress(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('AssociateAddress', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function allocateAddress(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('AllocateAddress', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function associateIamInstanceProfile(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('AssociateIamInstanceProfile', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function describeAccountAttributes(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DescribeAccountAttributes', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function disassociateIamInstanceProfile(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DisassociateIamInstanceProfile', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function replaceIamInstanceProfileAssociation(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('ReplaceIamInstanceProfileAssociation', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function describeIamInstanceProfileAssociations(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DescribeIamInstanceProfileAssociations', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function authorizeSecurityGroupIngress(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('AuthorizeSecurityGroupIngress', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function authorizeSecurityGroupEgress(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('AuthorizeSecurityGroupEgress', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function createImage(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('CreateImage', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function createKeyPair(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('CreateKeyPair', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function createNetworkInterface(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('CreateNetworkInterface', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function createTags(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('CreateTags', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteTags(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DeleteTags', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function createVolume(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('CreateVolume', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function createSnapshot(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('CreateSnapshot', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function createSecurityGroup(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('CreateSecurityGroup', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function deregisterImage(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DeregisterImage', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function describeInstances(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DescribeInstances', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function describeInstanceAttribute(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DescribeInstanceAttribute', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function describeImages(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DescribeImages', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function describeSecurityGroups(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DescribeSecurityGroups', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function describeNetworkInterfaces(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DescribeNetworkInterfaces', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function describeAddresses(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DescribeAddresses', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function describeSnapshots(array $params = []) {
    $params += $this->getDefaultParameters();
    $params['RestorableByUserIds'] = [$this->cloudConfigPluginManager->loadConfigEntity()->get('field_account_id')->value];
    $results = $this->execute('DescribeSnapshots', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function describeKeyPairs(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DescribeKeyPairs', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function describeVolumes(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DescribeVolumes', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function describeAvailabilityZones(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DescribeAvailabilityZones', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function describeVpcs(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DescribeVpcs', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function describeSubnets(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DescribeSubnets', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function getRegions() {
    $regions = EndpointProvider::defaultProvider()
      ->getPartition($region = '', 'ec2')['regions'];

    foreach ($regions as $region => $region_name) {
      $item[$region] = $region_name['description'];
    }

    return $item;
  }

  /**
   * {@inheritdoc}
   */
  public function getEndpointUrls() {
    // The $endpoints will be an array like ['us-east-1' => [], 'us-east-2' =>
    // [], ...].
    $endpoints = EndpointProvider::defaultProvider()
      ->getPartition($region = '', 'ec2')['services']['ec2']['endpoints'];

    $urls = [];
    foreach ($endpoints as $endpoint => $item) {
      $url = "https://ec2.$endpoint.amazonaws.com";
      $urls[$endpoint] = $url;
    }

    return $urls;
  }

  /**
   * {@inheritdoc}
   */
  public function importKeyPair(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('ImportKeyPair', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function terminateInstance(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('TerminateInstances', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteSecurityGroup(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DeleteSecurityGroup', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteNetworkInterface(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DeleteNetworkInterface', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function disassociateAddress(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DisassociateAddress', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function releaseAddress(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('ReleaseAddress', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteKeyPair(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DeleteKeyPair', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteVolume(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DeleteVolume', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function attachVolume(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('AttachVolume', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function detachVolume(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DetachVolume', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteSnapshot(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('DeleteSnapshot', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function revokeSecurityGroupIngress(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('RevokeSecurityGroupIngress', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function revokeSecurityGroupEgress(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('RevokeSecurityGroupEgress', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function runInstances(array $params = [], array $tags = []) {
    $params += $this->getDefaultParameters();

    // Add meta tags to identify where the instance was launched from.
    $params['TagSpecifications'] = [
      [
        'ResourceType' => 'instance',
        'Tags' => [
          [
            'Key' => 'cloud_launch_origin',
            'Value' => \Drupal::request()->getHost(),
          ],
          [
            'Key' => 'cloud_launch_software',
            'Value' => 'Drupal 8 Cloud Orchestrator',
          ],
          [
            'Key' => 'cloud_launched_by',
            'Value' => $this->currentUser->getAccountName(),
          ],
          [
            'Key' => 'cloud_launched_by_uid',
            'Value' => $this->currentUser->id(),
          ],
        ],
      ],
    ];

    // If there are tags, add them to the Tags array.
    foreach ($tags as $tag) {
      $params['TagSpecifications'][0]['Tags'][] = $tag;
    }

    $results = $this->execute('RunInstances', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function stopInstances(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('StopInstances', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function startInstances(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('StartInstances', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function modifyInstanceAttribute(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('ModifyInstanceAttribute', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function rebootInstances(array $params = []) {
    $params += $this->getDefaultParameters();
    $results = $this->execute('RebootInstances', $params);
    return $results;
  }

  /**
   * {@inheritdoc}
   */
  public function updateInstances(array $params = [], $clear = TRUE) {
    $timestamp = $this->getTimestamp();

    $updated = FALSE;
    $entity_type = 'aws_cloud_instance';

    // Call the api and get all instances.
    $result = $this->describeInstances($params);
    if ($result != NULL) {

      $all_instances = $this->loadAllEntities($entity_type);
      $stale = [];
      // Make it easier to lookup the instances by setting up
      // the array with the instance_id.
      foreach ($all_instances as $instance) {
        $stale[$instance->getInstanceId()] = $instance;
      }

      // Get instance IAM roles associated to instances.
      $instance_iam_roles = [];
      $associations_result = $this->describeIamInstanceProfileAssociations();
      foreach ($associations_result['IamInstanceProfileAssociations'] as $association) {
        $instance_iam_roles[$association['InstanceId']]
          = $association['IamInstanceProfile']['Arn'];
      }

      // Loop through the reservations and store each one as an Instance entity.
      foreach ($result['Reservations'] as $reservation) {

        foreach ($reservation['Instances'] as $instance) {
          // Keep track of instances that do not exist anymore
          // delete them after saving the rest of the instances.
          if (isset($stale[$instance['InstanceId']])) {
            unset($stale[$instance['InstanceId']]);
          }

          $instanceName = '';
          $user_id = 0;
          $termination_timestamp = NULL;
          $schedule = '';
          $tags = [];
          foreach ($instance['Tags'] as $tag) {
            if ($tag['Key'] == 'Name') {
              $instanceName = $tag['Value'];
            }
            if ($tag['Key'] == 'cloud_launched_by_uid') {
              $user_id = $tag['Value'];
            }
            if ($tag['Key'] == 'cloud_termination_timestamp') {
              if ($tag['Value'] != '') {
                $termination_timestamp = (int) $tag['Value'];
              }
            }
            if ($tag['Key'] == $this->configFactory->get('aws_cloud.settings')->get('aws_cloud_scheduler_tag')) {
              $schedule = $tag['Value'];
            }

            $tags[] = ['tag_key' => $tag['Key'], 'tag_value' => $tag['Value']];
          }

          usort($tags, function ($a, $b) {
            if ($a['tag_key'] == 'Name') {
              return -1;
            }

            if ($b['tag_key'] == 'Name') {
              return 1;
            }

            return strcmp($a['tag_key'], $b['tag_key']);
          });

          // Default to instance_id.
          if (empty($instanceName)) {
            $instanceName = $instance['InstanceId'];
          }

          $security_groups = [];
          foreach ($instance['SecurityGroups'] as $security_group) {
            $security_groups[] = $security_group['GroupName'];
          }

          // Termination protection.
          $attribute_result = $this->describeInstanceAttribute([
            'InstanceId' => $instance['InstanceId'],
            'Attribute' => 'disableApiTermination',
          ]);
          $termination_protection = $attribute_result['DisableApiTermination']['Value'];

          // Instance IAM roles.
          $iam_role = NULL;
          if (isset($instance_iam_roles[$instance['InstanceId']])) {
            $iam_role = $instance_iam_roles[$instance['InstanceId']];
          }

          // Use NetworkInterface to look up private ips.  In EC2-VPC,
          // an instance can have more than one private ip.
          $network_interfaces = [];
          $private_ips = FALSE;

          if (isset($instance['NetworkInterfaces'])) {
            $private_ips = $this->getPrivateIps($instance['NetworkInterfaces']);
            foreach ($instance['NetworkInterfaces'] as $interface) {
              $network_interfaces[] = $interface['NetworkInterfaceId'];
            }
          }

          $entity_id = $this->getEntityId($entity_type, 'instance_id', $instance['InstanceId']);

          // Skip if $entity already exists, by updating 'refreshed' time.
          if (!empty($entity_id)) {
            /* @var \Drupal\aws_cloud\Entity\Ec2\Instance $entity */
            $entity = Instance::load($entity_id);
            $entity->setName($instanceName);
            $entity->setInstanceState($instance['State']['Name']);

            // Set attributes that are available when system starts up.
            $public_ip = NULL;

            if ($private_ips != FALSE) {
              $entity->setPrivateIps($private_ips);
            }
            if (isset($instance['PublicIpAddress'])) {
              $public_ip = $instance['PublicIpAddress'];
            }
            if (isset($instance['PublicDnsName'])) {
              $entity->setPublicDns($instance['PublicDnsName']);
            }
            if (isset($instance['PrivateDnsName'])) {
              $entity->setPrivateDns($instance['PrivateDnsName']);
            }

            $entity->setPublicIp($public_ip);
            $entity->setSecurityGroups(implode(', ', $security_groups));
            $entity->setInstanceType($instance['InstanceType']);
            $entity->setRefreshed($timestamp);
            $entity->setLaunchTime($instance['LaunchTime']->__toString());
            $entity->setTerminationTimestamp($termination_timestamp);
            $entity->setTerminationProtection($termination_protection);
            $entity->setSchedule($schedule);
            $entity->setTags($tags);
            $entity->setIamRole($iam_role);
            $entity->setNetworkInterfaces($network_interfaces);

            if ($user_id != 0) {
              $entity->setOwnerById($user_id);
            }
            $entity->save();
            continue;
          }

          $entity = Instance::create([
            'cloud_context' => $this->cloudContext,
            'name' => !empty($instanceName) ? $instanceName : $instance['InstanceId'],
            'account_id' => $reservation['OwnerId'],
            'security_groups' => implode(', ', $security_groups),
            'instance_id' => $instance['InstanceId'],
            'instance_type' => $instance['InstanceType'],
            'availability_zone' => $instance['Placement']['AvailabilityZone'],
            'tenancy' => $instance['Placement']['Tenancy'],
            'instance_state' => $instance['State']['Name'],
            'public_dns' => $instance['PublicDnsName'],
            'public_ip' => isset($instance['PublicIpAddress']) ? $instance['PublicIpAddress'] : NULL,
            'private_dns' => $instance['PrivateDnsName'],
            'key_pair_name' => $instance['KeyName'],
            'is_monitoring' => $instance['Monitoring']['State'],
            'vpc_id' => $instance['VpcId'],
            'subnet_id' => $instance['SubnetId'],
            'source_dest_check' => $instance['SourceDestCheck'],
            'ebs_optimized' => $instance['EbsOptimized'],
            'root_device_type' => $instance['RootDeviceType'],
            'root_device' => $instance['RootDeviceName'],
            'image_id' => $instance['ImageId'],
            'placement_group' => $instance['Placement']['GroupName'],
            'virtualization' => $instance['VirtualizationType'],
            'reservation' => $reservation['ReservationId'],
            'ami_launch_index' => $instance['AmiLaunchIndex'],
            'host_id' => isset($instance['Placement']['HostId']) ? $instance['Placement']['HostId'] : NULL,
            'affinity' => isset($instance['Placement']['Affinity']) ? $instance['Placement']['Affinity'] : NULL,
            'state_transition_reason' => $instance['StateTransitionReason'],
            'instance_lock' => FALSE,
            'launch_time' => $instance['LaunchTime']->__toString(),
            'created' => strtotime($instance['LaunchTime']->__toString()),
            'changed' => $timestamp,
            'refreshed' => $timestamp,
            'user_id' => $user_id,
            'termination_timestamp' => $termination_timestamp,
            'termination_protection' => $termination_protection,
            'schedule' => $schedule,
            'tags' => $tags,
            'iam_role' => $iam_role,
          ]);

          if ($private_ips != FALSE) {
            $entity->setPrivateIps($private_ips);
          }
          $entity->setNetworkInterfaces($network_interfaces);
          $entity->save();
        }
      }

      if (count($stale) && $clear == TRUE) {
        $this->deleteStaleEntities($entity_type, $stale);
      }

      $updated = TRUE;
    }
    return $updated;
  }

  /**
   * {@inheritdoc}
   */
  public function updateImages(array $params = [], $clear = FALSE) {
    $timestamp = $this->getTimestamp();

    $updated = FALSE;
    $entity_type = 'aws_cloud_image';

    // Load all entities by cloud_context.
    $image_entities = $this->entityTypeManager->getStorage($entity_type)->loadByProperties(
      ['cloud_context' => $this->cloudContext]
    );

    $result = $this->describeImages($params);
    if ($result != NULL) {
      $stale = [];
      // Make it easier to lookup the images by setting up
      // the array with the image_id.
      foreach ($image_entities as $image) {
        $stale[$image->getImageId()] = $image;
      }

      $images = $result['Images'];
      foreach ($images as $image) {
        // Keep track of images that do not exist anymore
        // delete them after saving the rest of the images.
        if (isset($stale[$image['ImageId']])) {
          unset($stale[$image['ImageId']]);
        }

        $block_devices = [];
        foreach ($image['BlockDeviceMappings'] as $block_device) {
          $block_devices[] = $block_device['DeviceName'];
        }

        $user_id = $this->getUidTagValue($image, 'image_created_by_uid');

        $entity_id = $this->getEntityId($entity_type, 'image_id', $image['ImageId']);

        // Skip if $entity already exists, by updating 'refreshed' time.
        if (!empty($entity_id)) {
          $entity = Image::load($entity_id);
          $entity->setRefreshed($timestamp);
          $entity->setVisibility($image['Public']);
          $entity->setStatus($image['State']);

          if ($user_id != 0) {
            $entity->setOwnerById($user_id);
          }

          $entity->save();
          continue;
        }
        $entity = Image::create([
          'cloud_context' => $this->cloudContext,
          'image_id' => $image['ImageId'],
          'account_id' => $image['OwnerId'],
          'architecture' => $image['Architecture'],
          'virtualization_type' => $image['VirtualizationType'],
          'root_device_type' => $image['RootDeviceType'],
          'root_device_name' => $image['RootDeviceName'],
          'ami_name' => $image['Name'],
          'name' => $image['Name'],
          'kernel_id' => isset($image['KernelId']) ? $image['KernelId'] : '',
          'ramdisk_id' => isset($image['RamdiskId']) ? $image['RamdiskId'] : '',
          'image_type' => $image['ImageType'],
          'product_code' => isset($image['ProductCodes']) ? implode(',', array_column($image['ProductCodes'], 'ProductCode')) : '',
          'source' => $image['ImageLocation'],
          'state_reason' => isset($image['StateReason']) ? $image['StateReason']['Message'] : '',
          'platform' => isset($image['Platform']) ? $image['Platform'] : '',
          'description' => isset($image['Description']) ? $image['Description'] : '',
          'visibility' => $image['Public'],
          'block_devices' => implode(', ', $block_devices),
          'status' => $image['State'],
          'created' => strtotime($image['CreationDate']),
          'changed' => $timestamp,
          'refreshed' => $timestamp,
          'user_id' => $user_id,
        ]);
        $entity->save();
      }

      if (count($stale) && $clear == TRUE) {
        $this->deleteStaleEntities($entity_type, $stale);
      }
      $updated = count($images);
    }
    return $updated;
  }

  /**
   * {@inheritdoc}
   */
  public function updateSecurityGroups(array $params = [], $clear = TRUE) {
    $timestamp = $this->getTimestamp();
    $updated = FALSE;
    $entity_type = 'aws_cloud_security_group';

    $result = $this->describeSecurityGroups($params);
    if ($result != NULL) {
      $all_groups = $this->loadAllEntities($entity_type);
      $stale = [];
      // Make it easier to lookup the groups by setting up
      // the array with the group_id.
      foreach ($all_groups as $group) {
        $stale[$group->getGroupId()] = $group;
      }

      $security_groups = $result['SecurityGroups'];
      foreach ($security_groups as $security_group) {

        // Keep track of instances that do not exist anymore
        // delete them after saving the rest of the instances.
        if (isset($stale[$security_group['GroupId']])) {
          unset($stale[$security_group['GroupId']]);
        }

        $entity_id = $this->getEntityId($entity_type, 'group_id', $security_group['GroupId']);

        // Skip if $entity already exists, by updating 'refreshed' time.
        if (!empty($entity_id)) {
          /* @var \Drupal\aws_cloud\Entity\Ec2\SecurityGroup $entity */
          $entity = SecurityGroup::load($entity_id);
          $entity->setRefreshed($timestamp);
        }
        else {
          // Create a brand new SecurityGroup entity.
          $entity = SecurityGroup::create([
            'cloud_context' => $this->cloudContext,
            'name' => !empty($security_group['GroupName']) ? $security_group['GroupName'] : $security_group['GroupId'],
            'group_id' => $security_group['GroupId'],
            'group_name' => $security_group['GroupName'],
            'group_description' => $security_group['Description'],
            'vpc_id' => $security_group['VpcId'],
            'account_id' => $security_group['OwnerId'],
            'created' => $timestamp,
            'changed' => $timestamp,
            'refreshed' => $timestamp,
          ]);
        }

        if (isset($security_group['VpcId']) && !empty($security_group['VpcId'])) {
          // Check if VPC is default.  This involves another API call.
          $vpcs = $this->describeVpcs([
            'VpcIds' => [$security_group['VpcId']],
          ]);
          if ($vpcs['Vpcs']) {
            $default = $vpcs['Vpcs'][0]['IsDefault'];
            $entity->setDefaultVpc($default);
          }
        }

        // Setup the Inbound permissions.
        if (isset($security_group['IpPermissions'])) {
          $this->setupIpPermissions($entity, 'ip_permission', $security_group['IpPermissions']);
        }

        // Setup outbound permissions.
        if (isset($security_group['VpcId']) && isset($security_group['IpPermissionsEgress'])) {
          $this->setupIpPermissions($entity, 'outbound_permission', $security_group['IpPermissionsEgress']);
        }

        $entity->save();
      }

      if (count($stale) && $clear == TRUE) {
        $this->deleteStaleEntities($entity_type, $stale);
      }

      $updated = TRUE;
    }
    return $updated;
  }

  /**
   * Setup IP Permisssions.
   *
   * @param \Drupal\aws_cloud\Entity\Ec2\SecurityGroup $security_group
   *   The security group entity.
   * @param string $field
   *   Field to used for lookup.
   * @param array $ec2_permissions
   *   Permissions array from Ec2.
   */
  private function setupIpPermissions(SecurityGroup &$security_group, $field, array $ec2_permissions) {
    // Permissions are always overwritten with the latest from
    // EC2.  The reason is that there is no way to guarantee a 1
    // to 1 mapping from the $security_group['IpPermissions'] array.
    // There is no ip permission id coming back from EC2.
    // Clear out all items before re-adding them.
    $i = 0;
    while ($i <= $security_group->$field->count()) {
      if ($security_group->$field->get($i)) {
        $security_group->$field->removeItem($i);
      }
      $i++;
    }

    // Setup all permission objects.
    $count = 0;
    foreach ($ec2_permissions as $permissions) {
      $permission_objects = $this->setupIpPermissionObject($permissions);
      // Loop through the permission objects and add them to the
      // security group ip_permission field.
      foreach ($permission_objects as $permission) {
        $security_group->$field->set($count, $permission);
        $count++;
      }
    }
  }

  /**
   * Setup the ipermission field given the inbound security group array.
   *
   * The array comes from DescribeSecurityGroup EC2 api call.
   *
   * @param array $ec2_permission
   *   An array object of Ec2 permission.
   *
   * @return array
   *   An array of \Drupal\Core\Field\FieldItemInterface.
   */
  private function setupIpPermissionObject(array $ec2_permission) {
    $ip_permissions = [];

    // Get the field definition for an IpPermission object.
    $definition = $this->entityFieldManager->getBaseFieldDefinitions('aws_cloud_security_group');

    // Setup the more global attributes.
    $from_port = isset($ec2_permission['FromPort']) ? "{$ec2_permission['FromPort']}" : NULL;
    $to_port = isset($ec2_permission['ToPort']) ? "{$ec2_permission['ToPort']}" : NULL;
    $ip_protocol = $ec2_permission['IpProtocol'];

    // To keep things consistent, if ip_protocol is -1,
    // set from_port and to_port as 0-65535.
    if ($ip_protocol == -1) {
      $from_port = "0";
      $to_port = "65535";
    }

    if (isset($ec2_permission['IpRanges']) && count($ec2_permission['IpRanges'])) {
      // Create a IPv4 permission object.
      foreach ($ec2_permission['IpRanges'] as $ip_range) {
        $ip_range_permission = $this->fieldTypePluginManager->createInstance('ip_permission', [
          'field_definition' => $definition['ip_permission'],
          'parent' => NULL,
          'name' => NULL,
        ]);
        // Source is an internal identifier.  Doesn't come from EC2.
        $ip_range_permission->source = 'ip4';
        $ip_range_permission->cidr_ip = $ip_range['CidrIp'];

        $ip_range_permission->from_port = $from_port;
        $ip_range_permission->to_port = $to_port;
        $ip_range_permission->ip_protocol = $ip_protocol;

        $ip_permissions[] = $ip_range_permission;
      }
    }

    if (isset($ec2_permission['Ipv6Ranges']) && count($ec2_permission['Ipv6Ranges'])) {
      // Create IPv6 permissions object.
      foreach ($ec2_permission['Ipv6Ranges'] as $ip_range) {
        $ip_v6_permission = $this->fieldTypePluginManager->createInstance('ip_permission', [
          'field_definition' => $definition['ip_permission'],
          'parent' => NULL,
          'name' => NULL,
        ]);
        // Source is an internal identifier.  Doesn't come from EC2.
        $ip_v6_permission->source = 'ip6';
        $ip_v6_permission->cidr_ip_v6 = $ip_range['CidrIpv6'];
        $ip_v6_permission->from_port = $from_port;
        $ip_v6_permission->to_port = $to_port;
        $ip_v6_permission->ip_protocol = $ip_protocol;

        $ip_permissions[] = $ip_v6_permission;
      }
    }

    if (isset($ec2_permission['UserIdGroupPairs']) && count($ec2_permission['UserIdGroupPairs'])) {
      // Create Group permissions object.
      foreach ($ec2_permission['UserIdGroupPairs'] as $group) {
        $group_permission = $this->fieldTypePluginManager->createInstance('ip_permission', [
          'field_definition' => $definition['ip_permission'],
          'parent' => NULL,
          'name' => NULL,
        ]);
        // Source is an internal identifier.  Doesn't come from EC2.
        $group_permission->source = 'group';
        $group_permission->group_id = isset($group['GroupId']) ? $group['GroupId'] : NULL;
        $group_permission->group_name = isset($group['GroupName']) ? $group['GroupName'] : NULL;
        $group_permission->user_id = isset($group['UserId']) ? $group['UserId'] : NULL;
        $group_permission->peering_status = isset($group['PeeringStatus']) ? $group['PeeringStatus'] : NULL;
        $group_permission->vpc_id = isset($group['VpcId']) ? $group['VpcId'] : NULL;
        $group_permission->peering_connection_id = isset($group['VpcPeeringConnectionId']) ? $group['VpcPeeringConnectionId'] : NULL;

        $group_permission->from_port = $from_port;
        $group_permission->to_port = $to_port;
        $group_permission->ip_protocol = $ip_protocol;
        $ip_permissions[] = $group_permission;
      }
    }
    return $ip_permissions;
  }

  /**
   * {@inheritdoc}
   */
  public function updateNetworkInterfaces() {
    $timestamp = $this->getTimestamp();
    $updated = FALSE;
    $entity_type = 'aws_cloud_network_interface';

    $result = $this->describeNetworkInterfaces();
    if ($result != NULL) {
      $all_interfaces = $this->loadAllEntities($entity_type);
      $stale = [];
      // Make it easier to lookup the groups by setting up
      // the array with the group_id.
      foreach ($all_interfaces as $interface) {
        $stale[$interface->getNetworkInterfaceId()] = $interface;
      }

      $network_interfaces = $result['NetworkInterfaces'];
      foreach ($network_interfaces as $network_interface) {

        // Keep track of network interfaces that do not exist anymore
        // delete them after saving the rest of the network interfaces.
        if (isset($stale[$network_interface['NetworkInterfaceId']])) {
          unset($stale[$network_interface['NetworkInterfaceId']]);
        }

        // Set up the primary and secondary private ip addresses.
        // Setup the allocation_ids.  The allocation_ids are used during Elastic
        // Ip assignment.
        $primary_private_ip = NULL;
        $secondary_private_ip = NULL;
        $primary_association_id = NULL;
        $secondary_association_id = NULL;
        $public_ips = NULL;

        foreach ($network_interface['PrivateIpAddresses'] as $private_ip_address) {
          if ($private_ip_address['Primary'] == TRUE) {
            $primary_private_ip = $private_ip_address['PrivateIpAddress'];
            if (isset($private_ip_address['Association'])) {
              if (!empty($private_ip_address['Association']['AssociationId'])) {
                $primary_association_id = $private_ip_address['Association']['AssociationId'];
              }
              if (!empty($private_ip_address['Association']['PublicIp'])) {
                $public_ips[] = $private_ip_address['Association']['PublicIp'];
              }
            }
          }
          else {
            $secondary_private_ip = $private_ip_address['PrivateIpAddress'];
            if (isset($private_ip_address['Association'])) {
              if (!empty($private_ip_address['Association']['AssociationId'])) {
                $secondary_association_id = $private_ip_address['Association']['AssociationId'];
              }
              if (!empty($private_ip_address['Association']['PublicIp'])) {
                $public_ips[] = $private_ip_address['Association']['PublicIp'];
              }
            }
          }
        }

        $security_groups = [];
        foreach ($network_interface['Groups'] as $security_group) {
          $security_groups[] = $security_group['GroupName'];
        }

        $entity_id = $this->getEntityId($entity_type, 'network_interface_id', $network_interface['NetworkInterfaceId']);

        // Skip if $entity already exists, by updating 'refreshed' time.
        if (!empty($entity_id)) {
          /* @var \Drupal\aws_cloud\Entity\Ec2\NetworkInterface $entity */
          $entity = NetworkInterface::load($entity_id);
          $entity->setRefreshed($timestamp);
          $entity->setPrimaryPrivateIp($primary_private_ip);
          $entity->setSecondaryPrivateIp($secondary_private_ip);
          $entity->setAssociationId($primary_association_id);
          $entity->setSecondaryAssociationId($secondary_association_id);
          if ($public_ips != NULL) {
            $public_ips = implode(', ', $public_ips);
          }

          $entity->setPublicIps($public_ips);
          $entity->save();

          continue;
        }

        $entity = NetworkInterface::create([
          'cloud_context' => $this->cloudContext,
          'name' => $network_interface['NetworkInterfaceId'],
          'network_interface_id' => $network_interface['NetworkInterfaceId'],
          'vpc_id' => $network_interface['VpcId'],
          'mac_address' => $network_interface['MacAddress'],
          'security_groups' => implode(', ', $security_groups),
          'status' => $network_interface['Status'],
          'private_dns' => $network_interface['PrivateDnsName'],
          'primary_private_ip' => $primary_private_ip,
          'secondary_private_ips' => $secondary_private_ip,
          'attachment_id' => $network_interface['Attachment']['AttachmentId'],
          'attachment_owner' => $network_interface['Attachment']['InstanceOwnerId'],
          'attachment_status' => $network_interface['Attachment']['Status'],
          'account_id' => $network_interface['OwnerId'],
          'association_id' => $primary_association_id,
          'secondary_association_id' => $secondary_association_id,
          'subnet_id' => $network_interface['SubnetId'],
          'description' => $network_interface['Description'],
          'public_ips' => $public_ips,
          'source_dest_check' => $network_interface['SourceDestCheck'],
          'instance_id' => $network_interface['Attachment']['InstanceId'],
          'device_index' => $network_interface['Attachment']['DeviceIndex'],
          'delete_on_termination' => $network_interface['Attachment']['DeleteOnTermination'],
          'allocation_id' => $network_interface['Association']['AllocationId'],
          'created' => $timestamp,
          'changed' => $timestamp,
          'refreshed' => $timestamp,
        ]);
        $entity->save();
      }

      if (count($stale)) {
        $this->deleteStaleEntities($entity_type, $stale);
      }

      $updated = TRUE;
    }
    return $updated;

  }

  /**
   * {@inheritdoc}
   */
  public function updateElasticIp() {
    $timestamp = $this->getTimestamp();
    $updated = FALSE;
    $entity_type = 'aws_cloud_elastic_ip';

    $result = $this->describeAddresses();

    if ($result != NULL) {
      $all_ips = $this->loadAllEntities($entity_type);
      $stale = [];
      // Make it easier to lookup the groups by setting up
      // the array with the group_id.
      foreach ($all_ips as $ip) {
        $stale[$ip->getPublicIp()] = $ip;
      }

      $elastic_ips = $result['Addresses'];
      foreach ($elastic_ips as $elastic_ip) {
        // Keep track of Ips that do not exist anymore
        // delete them after saving the rest of the Ips.
        if (isset($stale[$elastic_ip['PublicIp']])) {
          unset($stale[$elastic_ip['PublicIp']]);
        }
        $elastic_ip_name = '';
        if (isset($elastic_ip['Tags'])) {
          foreach ($elastic_ip['Tags'] as $tag) {
            if ($tag['Key'] == 'Name') {
              $elastic_ip_name = $tag['Value'];
            }
          }
        }
        // Default to instance_id.
        if (empty($elastic_ip_name)) {
          $elastic_ip_name = $elastic_ip['PublicIp'];
        }

        $entity_id = $this->getEntityId($entity_type, 'public_ip', $elastic_ip['PublicIp']);

        // Skip if $entity already exists, by updating 'refreshed' time.
        if (!empty($entity_id)) {
          $entity = ElasticIp::load($entity_id);

          // Update fields.
          $entity->setName($elastic_ip_name);
          $entity->setInstanceId(!empty($elastic_ip['InstanceId']) ? $elastic_ip['InstanceId'] : '');
          $entity->setNetworkInterfaceId(!empty($elastic_ip['NetworkInterfaceId']) ? $elastic_ip['NetworkInterfaceId'] : '');
          $entity->setPrivateIpAddress(!empty($elastic_ip['PrivateIpAddress']) ? $elastic_ip['PrivateIpAddress'] : '');
          $entity->setNetworkInterfaceOwner(!empty($elastic_ip['NetworkInterfaceOwnerId']) ? $elastic_ip['NetworkInterfaceOwnerId'] : '');
          $entity->setAllocationId(!empty($elastic_ip['AllocationId']) ? $elastic_ip['AllocationId'] : '');
          $entity->setAssociationId(!empty($elastic_ip['AssociationId']) ? $elastic_ip['AssociationId'] : '');
          $entity->setDomain(!empty($elastic_ip['Domain']) ? $elastic_ip['Domain'] : '');

          $entity->setRefreshed($timestamp);
          $entity->save();
          continue;
        }

        $entity = ElasticIp::create([
          'cloud_context' => $this->cloudContext,
          'name' => $elastic_ip_name,
          'public_ip' => $elastic_ip['PublicIp'],
          'instance_id' => !empty($elastic_ip['InstanceId']) ? $elastic_ip['InstanceId'] : '',
          'network_interface_id' => !empty($elastic_ip['NetworkInterfaceId']) ? $elastic_ip['NetworkInterfaceId'] : '',
          'private_ip_address' => !empty($elastic_ip['PrivateIpAddress']) ? $elastic_ip['PrivateIpAddress'] : '',
          'network_interface_owner' => !empty($elastic_ip['NetworkInterfaceOwnerId']) ? $elastic_ip['NetworkInterfaceOwnerId'] : '',
          'allocation_id' => !empty($elastic_ip['AllocationId']) ? $elastic_ip['AllocationId'] : '',
          'association_id' => !empty($elastic_ip['AssociationId']) ? $elastic_ip['AssociationId'] : '',
          'domain' => !empty($elastic_ip['Domain']) ? $elastic_ip['Domain'] : '',
          'created' => $timestamp,
          'changed' => $timestamp,
          'refreshed' => $timestamp,
        ]);
        $entity->save();
      }

      if (count($stale)) {
        $this->deleteStaleEntities($entity_type, $stale);
      }
      $updated = TRUE;
    }
    return $updated;
  }

  /**
   * {@inheritdoc}
   */
  public function updateKeyPairs() {
    $timestamp = $this->getTimestamp();
    $updated = FALSE;

    $entity_type = 'aws_cloud_key_pair';

    $result = $this->describeKeyPairs();
    if ($result != NULL) {
      $all_keys = $this->loadAllEntities($entity_type);
      $stale = [];
      // Make it easier to lookup the groups by setting up
      // the array with the group_id.
      foreach ($all_keys as $key) {
        $stale[$key->getKeyPairName()] = $key;
      }

      $key_pairs = $result['KeyPairs'];
      foreach ($key_pairs as $key_pair) {

        // Keep track of key pair that do not exist anymore
        // delete them after saving the rest of the key pair.
        if (isset($stale[$key_pair['KeyName']])) {
          unset($stale[$key_pair['KeyName']]);
        }

        $entity_id = $this->getEntityId($entity_type, 'key_pair_name', $key_pair['KeyName']);

        // Skip if $entity already exists, by updating 'refreshed' time.
        if (!empty($entity_id)) {
          $entity = KeyPair::load($entity_id);
          $entity->setRefreshed($timestamp);
          $entity->save();
          continue;
        }

        $entity = KeyPair::create([
          // $cloud_context,.
          'cloud_context' => $this->cloudContext,
          'key_pair_name' => $key_pair['KeyName'],
          'key_fingerprint' => $key_pair['KeyFingerprint'],
          'created' => $timestamp,
          'changed' => $timestamp,
          'refreshed' => $timestamp,
        ]);
        $entity->save();
      }

      if (count($stale)) {
        $this->deleteStaleEntities($entity_type, $stale);
      }

      $updated = TRUE;
    }
    return $updated;
  }

  /**
   * {@inheritdoc}
   */
  public function updateVolumes() {
    $timestamp = $this->getTimestamp();
    $updated = FALSE;

    $entity_type = 'aws_cloud_volume';

    $result = $this->describeVolumes();

    if ($result != NULL) {
      $all_volumes = $this->loadAllEntities($entity_type);
      $stale = [];
      // Make it easier to lookup the groups by setting up
      // the array with the group_id.
      foreach ($all_volumes as $volume) {
        $stale[$volume->getVolumeId()] = $volume;
      }

      // Get names of snapshot.
      $snapshot_id_name_map = $this->getSnapshotIdNameMap($result['Volumes']);

      $volumes = $result['Volumes'];
      foreach ($volumes as $volume) {
        // Keep track of network interfaces that do not exist anymore
        // delete them after saving the rest of the network interfaces.
        if (isset($stale[$volume['VolumeId']])) {
          unset($stale[$volume['VolumeId']]);
        }

        $attachments = [];
        foreach ($volume['Attachments'] as $attachment) {
          $attachments[] = $attachment['InstanceId'];
        }

        $name = $this->getTagName($volume, $volume['VolumeId']);

        $entity_id = $this->getEntityId($entity_type, 'volume_id', $volume['VolumeId']);
        $user_id = $this->getUidTagValue($volume, 'volume_created_by_uid');

        if ($user_id == 0) {
          // Inherit the volume uid from the instance that launched it.
          if (count($attachments)) {
            $user_id = $this->getInstanceUid($attachments[0]);
          }
        }

        // Skip if $entity already exists, by updating 'refreshed' time.
        if (!empty($entity_id)) {
          /* @var \Drupal\aws_cloud\Entity\Ec2\Volume $entity */
          $entity = Volume::load($entity_id);
          $entity->setName($name);
          $entity->setRefreshed($timestamp);
          $entity->setState($volume['State']);
          $entity->setAttachmentInformation(implode(', ', $attachments));
          $entity->setCreated(strtotime($volume['CreateTime']));
          $entity->setSnapshotId($volume['SnapshotId']);
          $entity->setSnapshotName(empty($volume['SnapshotId'])
            ? ''
            : $snapshot_id_name_map[$volume['SnapshotId']]);
          if ($user_id != 0) {
            $entity->setOwnerById($user_id);
          }
          $entity->save();

          continue;
        }
        $entity = Volume::create([
          'cloud_context' => $this->cloudContext,
          'name' => $name,
          'volume_id' => $volume['VolumeId'],
          'size' => $volume['Size'],
          'state' => $volume['State'],
          'volume_status' => $volume['VirtualizationType'],
          'attachment_information' => implode(', ', $attachments),
          'volume_type' => $volume['VolumeType'],
          'iops' => $volume['Iops'],
          'snapshot_id' => $volume['SnapshotId'],
          'snapshot_name' => empty($volume['SnapshotId']) ? '' : $snapshot_id_name_map[$volume['SnapshotId']],
          'availability_zone' => $volume['AvailabilityZone'],
          'encrypted' => $volume['Encrypted'],
          'kms_key_id' => $volume['KmsKeyId'],
          'created' => strtotime($volume['CreateTime']),
          'changed' => $timestamp,
          'refreshed' => $timestamp,
          'user_id' => $user_id,
        ]);
        $entity->save();
      }

      if (count($stale)) {
        $this->deleteStaleEntities($entity_type, $stale);
      }

      $updated = TRUE;
    }

    return $updated;
  }

  /**
   * {@inheritdoc}
   */
  public function updateSnapshots() {
    $timestamp = $this->getTimestamp();
    $updated = FALSE;

    $entity_type = 'aws_cloud_snapshot';

    $result = $this->describeSnapshots();
    if ($result != NULL) {
      $all_snapshots = $this->loadAllEntities($entity_type);
      $stale = [];
      // Make it easier to lookup the snapshot by setting up
      // the array with the snapshot_id.
      foreach ($all_snapshots as $snapshot) {
        $stale[$snapshot->getSnapshotId()] = $snapshot;
      }

      $snapshots = $result['Snapshots'];
      foreach ($snapshots as $snapshot) {
        // Keep track of snapshot that do not exist anymore
        // delete them after saving the rest of the snapshots.
        if (isset($stale[$snapshot['SnapshotId']])) {
          unset($stale[$snapshot['SnapshotId']]);
        }

        $name = $this->getTagName($snapshot, $snapshot['SnapshotId']);
        $entity_id = $this->getEntityId($entity_type, 'snapshot_id', $snapshot['SnapshotId']);
        $user_id = $this->getUidTagValue($snapshot, 'snapshot_created_by_uid');

        // Skip if $entity already exists, by updating 'refreshed' time.
        if (!empty($entity_id)) {
          $entity = Snapshot::load($entity_id);
          $entity->setName($name);
          $entity->setStatus($snapshot['State']);
          $entity->setRefreshed($timestamp);
          if ($user_id != 0) {
            $entity->setOwnerById($user_id);
          }
          $entity->save();
          continue;
        }

        $entity = Snapshot::create([
          'cloud_context' => $this->cloudContext,
          'name' => $name,
          'snapshot_id' => $snapshot['SnapshotId'],
          'size' => $snapshot['VolumeSize'],
          'description' => $snapshot['Description'],
          'status' => $snapshot['State'],
          'volume_id' => $snapshot['VolumeId'],
          'progress' => $snapshot['Progress'],
          'encrypted' => $snapshot['Encrypted'] == FALSE ? 'Not Encrypted' : 'Encrypted',
          'kms_key_id' => $snapshot['KmsKeyId'],
          'account_id' => $snapshot['OwnerId'],
          'owner_aliases' => $snapshot['OwnerAlias'],
          'state_message' => $snapshot['StateMessage'],
          'created' => strtotime($snapshot['StartTime']),
          'changed' => $timestamp,
          'refreshed'  => $timestamp,
          'user_id' => $user_id,
        ]);
        $entity->save();
      }
      if (count($stale)) {
        $this->deleteStaleEntities($entity_type, $stale);
      }
      $updated = TRUE;
    }

    return $updated;
  }

  /**
   * {@inheritdoc}
   */
  public function getVpcs() {
    $vpcs = [];
    $results = $this->describeVpcs();
    foreach (array_column($results['Vpcs'], 'VpcId') as $key => $vpc) {
      $vpcs[$vpc] = $results['Vpcs'][$key]['CidrBlock'] . " ($vpc)";
    }
    return $vpcs;
  }

  /**
   * {@inheritdoc}
   */
  public function getAvailabilityZones() {
    $zones = [];
    $results = $this->describeAvailabilityZones();
    if ($results != NULL) {
      foreach (array_column($results['AvailabilityZones'], 'ZoneName') as $availability_zone) {
        $zones[$availability_zone] = $availability_zone;
      }
    }
    return $zones;
  }

  /**
   * {@inheritdoc}
   */
  public function getSupportedPlatforms() {
    $platforms = [];
    $results = $this->describeAccountAttributes([
      'AttributeNames' => [
        'supported-platforms',
      ],
    ]);
    if ($results != NULL) {
      foreach ($results['AccountAttributes'] as $attribute) {
        if ($attribute['AttributeName'] == 'supported-platforms') {
          foreach ($attribute['AttributeValues'] as $value) {
            $platforms[] = $value['AttributeValue'];
          }
        }
      }
    }
    return $platforms;
  }

  /**
   * {@inheritdoc}
   */
  public function clearAllEntities() {
    $timestamp = $this->getTimestamp();
    $this->clearEntities('aws_cloud_instance', $timestamp);
    $this->clearEntities('aws_cloud_security_group', $timestamp);
    $this->clearEntities('aws_cloud_image', $timestamp);
    $this->clearEntities('aws_cloud_network_interface', $timestamp);
    $this->clearEntities('aws_cloud_elastic_ip', $timestamp);
    $this->clearEntities('aws_cloud_key_pair', $timestamp);
    $this->clearEntities('aws_cloud_volume', $timestamp);
    $this->clearEntities('aws_cloud_snapshot', $timestamp);
  }

  /**
   * Helper method to get the current timestamp.
   *
   * @return int
   *   The current timestamp.
   */
  private function getTimestamp() {
    return time();
  }

  /**
   * Setup the default parameters that all API calls will need.
   *
   * @return array
   *   Array of default parameters.
   */
  private function getDefaultParameters() {
    return [
      'DryRun' => $this->dryRun,
    ];
  }

  /**
   * Helper method to delete entities.
   *
   * @param string $entity_type
   *   Entity Type.
   * @param array $entity_ids
   *   Array of entity ids.
   */
  private function deleteEntities($entity_type, array $entity_ids) {
    $entities = $this->entityTypeManager->getStorage($entity_type)->loadMultiple($entity_ids);
    $this->entityTypeManager->getStorage($entity_type)->delete($entities);
  }

  /**
   * Clear entities.
   *
   * @param string $entity_type
   *   Entity Type.
   * @param int $timestamp
   *   The timestamp for condition of refreshed time to clear entities.
   */
  private function clearEntities($entity_type, $timestamp) {
    $entity_ids = $this->entityTypeManager->getStorage($entity_type)->getQuery()
      ->condition('refreshed', $timestamp, '<')
      ->condition('cloud_context', $this->cloudContext)
      ->execute();
    if (count($entity_ids)) {
      $this->deleteEntities($entity_type, $entity_ids);
    }
  }

  /**
   * Helper method to load an entity using parameters.
   *
   * @param string $entity_type
   *   Entity Type.
   * @param string $id_field
   *   Entity id field.
   * @param string $id_value
   *   Entity id value.
   *
   * @return int
   *   Entity id.
   */
  private function getEntityId($entity_type, $id_field, $id_value) {
    $entities = $this->entityTypeManager->getStorage($entity_type)->getQuery()
      ->condition($id_field, $id_value)
      ->condition('cloud_context', $this->cloudContext)
      ->execute();
    return array_shift($entities);
  }

  /**
   * Helper method to load all entities of a given type.
   *
   * @param string $entity_type
   *   Entity type.
   *
   * @return \Drupal\Core\Entity\EntityInterface[]
   *   Array of entities.
   */
  private function loadAllEntities($entity_type) {
    return $this->entityTypeManager->getStorage($entity_type)->loadByProperties(
      ['cloud_context' => [$this->cloudContext]]
    );
  }

  /**
   * Helper method to delete stale entities.
   *
   * @param string $entity_type
   *   Entity type.
   * @param array $entities
   *   Array of entities.
   */
  private function deleteStaleEntities($entity_type, array $entities) {
    $this->entityTypeManager->getStorage($entity_type)->delete($entities);
  }

  /**
   * Helper method to get the map of snapshot ID and name.
   *
   * @param array $volumes
   *   Array of volumes.
   *
   * @return array
   *   Map of snapshots.
   */
  private function getSnapshotIdNameMap(array $volumes) {
    $snapshot_ids = array_filter(array_column($volumes, 'SnapshotId'));
    if (empty($snapshot_ids)) {
      return [];
    }

    $map = [];
    foreach ($snapshot_ids as $snapshot_id) {
      $map[$snapshot_id] = '';
    }

    $result = $this->describeSnapshots();
    foreach ($result['Snapshots'] as $snapshot) {
      $snapshot_id = $snapshot['SnapshotId'];
      if (!array_key_exists($snapshot_id, $map)) {
        continue;
      }

      $map[$snapshot_id] = $this->getTagName($snapshot, '');
    }

    return $map;
  }

  /**
   * Helper method to get the name of aws object.
   *
   * @param array $aws_obj
   *   Array of aws pbject.
   * @param string $default_value
   *   Default value of tag name.
   *
   * @return string
   *   Tag name.
   */
  private function getTagName(array $aws_obj, $default_value) {
    $name = $default_value;
    if (!isset($aws_obj['Tags'])) {
      return $name;
    }

    foreach ($aws_obj['Tags'] as $tag) {
      if ($tag['Key'] == 'Name') {
        $name = $tag['Value'];
        break;
      }
    }
    return $name;
  }

  /**
   * Helper function to loop the network interfaces.
   *
   * Also creates a comma delimited string of private ips. Function returns
   * false if no private ips found.
   *
   * @param array $network_interfaces
   *   Array of network interfaces from the EC2 DescribeInstance api.
   *
   * @return string|false
   *   Imploded string or FALSE if no private ips round.
   */
  private function getPrivateIps(array $network_interfaces) {
    $ip_string = FALSE;
    $private_ips = [];
    foreach ($network_interfaces as $interface) {
      $private_ips[] = $interface['PrivateIpAddress'];
    }
    if (count($private_ips)) {
      $ip_string = implode(', ', $private_ips);
    }
    return $ip_string;
  }

  /**
   * Helper function to parse drupal uid value out of the tags array.
   *
   * @param array $tags_array
   *   The tags array.
   * @param string $key
   *   The uid key.
   *
   * @return int
   *   Drupal uid.
   */
  private function getUidTagValue(array $tags_array, $key) {
    $user_id = 0;
    if (isset($tags_array['Tags'])) {
      foreach ($tags_array['Tags'] as $tag) {
        if ($tag['Key'] == $key) {
          $user_id = $tag['Value'];
          break;
        }
      }
    }
    return $user_id;
  }

  /**
   * Helper function to get an instance's uid.
   *
   * @param string $instance_id
   *   The instance_id to load.
   *
   * @return int
   *   The uid of the instance.
   */
  private function getInstanceUid($instance_id) {
    $user_id = 0;
    $instance = $this->entityTypeManager
      ->getStorage('aws_cloud_instance')
      ->loadByProperties([
        'instance_id' => $instance_id,
      ]);

    if (count($instance) > 0) {
      $instance = array_shift($instance);
      $user_id = $instance->getOwnerId();
    }
    return $user_id;
  }

}
